/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package sync4j.syncclient.blackberry.parser;

import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.Vector;

import net.rim.blackberry.api.pim.Event;
import net.rim.blackberry.api.pim.EventList;
import net.rim.device.api.ui.component.Dialog;
import net.rim.device.api.i18n.SimpleDateFormat;

import sync4j.syncclient.util.StaticDataHelper;

/**
 * Provides with the user interface
 * for sync events with SyncServer.
 *
 * @author: Fabio Maggi @ Funambol
 * $Id: XMLEventParser.java,v 1.0
 */

public class XMLEventParser implements EventParser{

    //------------------------------------------Private Data Constants

    private static final String SUMMARY          = "Subject"                    ;
    private static final String LOCATION         = "Location"                   ;
    private static final String NOTE             = "Body"                       ;
    private static final String START            = "Start"                      ;
    private static final String END              = "End"                        ;
    private static final String REVISION         = "Revision"                   ;
    private static final String ALARM            = "ReminderMinutesBeforeStart" ;
    private static final String CLASS            = "Class"                      ;

    private static final String START_MARKER     = "<appointment>"              ;
    private static final String END_MARKER       = "</appointment>"             ;

    private static final String DATETIME_FORMAT  = "yyyyMMdd'T'HHmmss"          ;
    private static final String UTC_TIMEZONE     = "UTC"                        ;

    //------------------------------------------Private Data menmbers

    private String[] tagArray = {
            SUMMARY      ,
            LOCATION     ,
            NOTE         ,
            START        ,
            END          ,
            REVISION     ,
            ALARM        ,
            CLASS
    };

    private EventList list   ;
    private Event     event  ;
    private boolean   modify ;

    /**
     * This is an initialization constructor for XMLEventParser class.
     *
     * @param EventList: Blackberry event list
     * @param Event: Blackberry event
     * @param boolean: to modify or not to modify
     * @return XMLEventParser
     */
    public XMLEventParser(EventList list, Event event, boolean modify) {

        this.list   = list   ;
        this.event  = event  ;
        this.modify = modify ;
    }

    /**
     * This method is invoked to obtain Event object from a event String.
     * if the modify flag is set to true it modifies the event.
     * if the modify flag is set to false it simply adds a event to blackberry calendar.
     *
     * @param String:String containing event information to be parsed.
     * @return Event: that has been added newly/modified in blackberry calendar.
     */
    public Event parseEvent(String eventString) {

        if(modify) {
            return modifyEvent (eventString);
        } else {
            return addEvent    (eventString);
        }

    }

    /**
     * This method is invoked to obtain Event object from a event String,
     * when a event is added to the blackberry event.This method
     * adds a new event and returns the reference of that object.
     *
     * @param String: String containing event information to be parsed.
     * @return Event: Event object that has been added newly.
     */
    public Event addEvent(String eventString) {

        Hashtable eventMap = buildMapFromXML(eventString);
        if (eventMap != null) {
            try {

                String field = getValue(eventMap, SUMMARY);
                if (isSupportedField(Event.SUMMARY, field)) {
                    event.addString(Event.SUMMARY, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, LOCATION);
                if (isSupportedField(Event.LOCATION, field)) {
                    event.addString(Event.LOCATION, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, NOTE);
                if (isSupportedField(Event.NOTE, field)) {
                    event.addString(Event.NOTE, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, START);
                if (field != null && field.length() > 0) {
                    long dateStart = getDate(field);
                    if (isSupportedField(Event.START, field)) {
                        event.addDate(Event.START, Event.ATTR_NONE, dateStart);
                    }
                }

                field = getValue(eventMap, END);
                if (field != null && field.length() > 0) {
                    long dateEnd = getDate(field);
                    if (isSupportedField(Event.END, field)) {
                        event.addDate(Event.END, Event.ATTR_NONE, dateEnd);
                    }
                }

                field = getValue(eventMap, REVISION);
                if (field != null && field.length() > 0) {
                    long dateRevision = getDate(field);
                    if (isSupportedField(Event.REVISION, field)) {
                        event.addDate(Event.REVISION, Event.ATTR_NONE, dateRevision);
                    }
                }

                field = getValue(eventMap, ALARM);
                if (isSupportedField(Event.ALARM, field)) {
                    event.addInt(Event.ALARM, Event.ATTR_NONE, Integer.parseInt(field));
                }

                field = getValue(eventMap, CLASS);
                if (isSupportedField(Event.CLASS, field)) {
                    event.addInt(Event.CLASS, Event.ATTR_NONE, Integer.parseInt(field));
                }

                return event;

            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }
        return null;
    }

    /**
     * This method is invoked to obtain Event object from a event String,
     * when a event is modified in the blackberry calendar. This method
     * modifies the object and returns the reference to that object.
     *
     * @param String: String containing event information to be parsed.
     * @return Event:Event that has been modified.
     */
    public Event modifyEvent(String eventString) {

        Hashtable eventMap = buildMapFromXML(eventString);
        if (eventMap != null) {
            try {

                String field = getValue(eventMap, SUMMARY);
                if (isSupportedAtrributedField(Event.SUMMARY, Event.ATTR_NONE, field) && event.countValues(Event.SUMMARY) > 0) {
                        event.setString(Event.SUMMARY, 0, Event.ATTR_NONE, field);
                } else {
                        event.addString(Event.SUMMARY, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, LOCATION);
                if (isSupportedAtrributedField(Event.LOCATION, Event.ATTR_NONE, field) && event.countValues(Event.LOCATION) > 0) {
                        event.setString(Event.LOCATION, 0, Event.ATTR_NONE, field);
                } else {
                        event.addString(Event.LOCATION, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, NOTE);

                if (isSupportedAtrributedField(Event.NOTE, Event.ATTR_NONE, field) && event.countValues(Event.NOTE) > 0) {
                        event.setString(Event.NOTE, 0, Event.ATTR_NONE, field);
                } else {
                        event.addString(Event.NOTE, Event.ATTR_NONE, field);
                }

                field = getValue(eventMap, START);
                if (field != null && field.length() > 0) {
                    long dateStart = getDate(field);
                    if (isSupportedAtrributedField(Event.START, Event.ATTR_NONE, field)) {
                            event.removeValue(Event.START, 0);
                            event.addDate(Event.START, Event.ATTR_NONE, dateStart);
                    }
                }

                field = getValue(eventMap, END);
                if (field != null && field.length() > 0) {
                    long dateEnd = getDate(field);
                    if (isSupportedAtrributedField(Event.END, Event.ATTR_NONE, field)) {
                            event.removeValue(Event.END, 0);
                            event.addDate(Event.END, Event.ATTR_NONE, dateEnd);
                    }
                }

                field = getValue(eventMap, REVISION);
                if (field != null && field.length() > 0) {
                    long dateRevision = getDate(field);
                    if (isSupportedAtrributedField(Event.END, Event.ATTR_NONE, field)) {
                            event.removeValue(Event.REVISION, 0);
                            event.addDate(Event.REVISION, Event.ATTR_NONE, dateRevision);
                    }
                }

                field = getValue(eventMap, ALARM);
                int alarmSecondsBefore = getAlarmSecondsBefore(field);
                if (isSupportedAtrributedField(Event.ALARM, Event.ATTR_NONE, field) && event.countValues(Event.ALARM) > 0) {
                        event.setInt(Event.ALARM, 0, Event.ATTR_NONE,alarmSecondsBefore);
                } else {
                        event.addInt(Event.ALARM, Event.ATTR_NONE, alarmSecondsBefore);
                }


                field = getValue(eventMap, CLASS);
                if (isSupportedAtrributedField(Event.CLASS, Event.ATTR_NONE, field)) {
                        event.removeValue(Event.ALARM, 0);
                        event.addInt(Event.CLASS, Event.ATTR_NONE, Integer.parseInt(field));
                }

                return event;
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }
        return null;
    }

    /**
     * This method appends all the event information passed to it
     * in the form of event object.
     *
     * @param Event: Blackberry event.
     * @return String: all event information in a string format.
     */
    public String toString(Event event) {

        StringBuffer eventBuffer = new StringBuffer();

        eventBuffer.append("<?xml version=\"1.0\" encoding=\""          +
                           System.getProperty("microedition.encoding")  +
                           "\"?>"                                       );

        eventBuffer.append("<calendar>\n");

        appendField           (eventBuffer      ,
                               event            ,
                               Event.SUMMARY    ,
                               SUMMARY          );

        appendField           (eventBuffer      ,
                               event            ,
                               Event.LOCATION   ,
                               LOCATION         );

        appendField           (eventBuffer      ,
                               event            ,
                               Event.NOTE       ,
                               NOTE             );

        appendFieldDate       (eventBuffer      ,
                               event            ,
                               Event.START      ,
                               START            );

        appendFieldDate       (eventBuffer      ,
                               event            ,
                               Event.END        ,
                               END              );

        appendFieldDate       (eventBuffer      ,
                               event            ,
                               Event.REVISION   ,
                               REVISION         );

        appendFieldAlarmMB    (eventBuffer      ,
                               event            ,
                               Event.ALARM      ,
                               ALARM            );

        appendFieldInt        (eventBuffer      ,
                               event            ,
                               Event.CLASS      ,
                               CLASS            );

        eventBuffer.append("</calendar>");

        return eventBuffer.toString();
    }

    //-----------------------------------------Private Methods

    /**
     * This method appends the event information String Type supported by blackberry to a
     * string buffer.
     *
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @param int: type of field
     * @param String: Name of field.
     * @return void.
     */
    private void appendField(StringBuffer eventBuffer, Event event, int field, String tag) {

        String value = null;

        if (list.isSupportedField(field)) {

            value = "";

            try {
                value = event.getString(field, 0);
            } catch (Exception e) {
                System.out.println("Not set field: " + tag);
            }

            appendToEvent(eventBuffer, tag, value);
        }
    }

    /**
     * This method appends the event information Date Type supported by blackberry to a
     * string buffer.
     *
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @param int: type of field
     * @param String: Name of field.
     * @return void.
     */
    private void appendFieldDate(StringBuffer eventBuffer, Event event, int field, String tag) {

        StaticDataHelper  sdh            = null ;

        long              value          = 0    ;
        Calendar          rightNow       = null ;
        String            formattedDate  = null ;
        SimpleDateFormat  formatter      = null ;

        long              offset         = 0    ;

        sdh = new StaticDataHelper();

        if (list.isSupportedField(field)) {

            try {
                value = event.getDate(field, 0);
            } catch (Exception e) {
                System.out.println("Not set field: " + tag);
            }

            if (sdh.enableAdjustDate()) {
                offset = getTimeZoneOffSet();
            } else {
                offset = 0;
            }

            rightNow = Calendar.getInstance();

            rightNow.setTime(new Date(value));

            formatter      = new SimpleDateFormat (DATETIME_FORMAT );
            formattedDate  = formatter.format     (new Date(rightNow.getTime().getTime() - offset));

            appendToEvent(eventBuffer, tag, formattedDate);
        }
    }

    /**
     * This method appends the event information int Type supported by blackberry to a
     * string buffer.
     *
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @param int: type of field
     * @param String: Name of field.
     * @return void.
     */
    private void appendFieldInt(StringBuffer eventBuffer, Event event, int field, String tag) {

        String value = null ;

        if (list.isSupportedField(field)) {

            value = "";

            try {
                value = String.valueOf(event.getInt(field, 0));
            } catch (Exception e) {
                System.out.println("Not set field: " + tag);
            }

            appendToEvent(eventBuffer, tag, value);
        }
    }

    /**
     * This method appends the event information int Type supported by blackberry to a
     * string buffer.
     *
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @param int: type of field
     * @param String: Name of field.
     * @return void.
     */
    private void appendFieldAlarmMB(StringBuffer eventBuffer, Event event, int field, String tag) {

        String  value = null ;

        if (list.isSupportedField(field)) {

            value = "";

            try {
                value = String.valueOf((event.getInt(field, 0))/60);
            } catch (Exception e) {
                System.out.println("Not set field: " + tag);
            }

            appendToEvent(eventBuffer, tag, value);
        }
    }

    /**
     * This method appends the event information supported by blackberry to a
     * string buffer.
     *
     * @param StringBuffer: StringBuffer containing event information.
     * @param String: name of tag associated with event information.
     * @param String: value corresponding to tag.
     * @return void
     */
    private void appendToEvent(StringBuffer eventBuffer, String tag, String value) {

        value = (value == null) ? "" : value;

        eventBuffer.append("<").append(tag).append(">").append(value);

        eventBuffer.append("</").append(tag).append(">\n");

    }

    /**
     * This method returns the value associated with a key in a hashtable
     * containing eventMap.
     *
     * @param Hashtable: eventMap
     * @param String: key to which associated value it to be found.
     * @return String: Value associated with the key
     */
    private String getValue(Hashtable eventMap, String key) {
        Object val = eventMap.get(key);
        String value = val == null ? "" : val.toString();
        return value.trim();
    }

    /**
     * This method checks if the field and attibute are supported by the event list
     * of blackberry address book
     *
     * @param int: field
     * @param int: attribute
     * @param String: value
     * @return boolean:returns true if both the field and attribute
     * associated with it r supported
     */
    private boolean isSupportedAtrributedField(int field, int attribute, String value) {
        return isSupportedField(field, value) && list.isSupportedAttribute(field, attribute);
    }

    /**
     * This method checks if the type of field is supported by the event list
     * of blackberry address book.
     *
     * @param int: type
     * @param String: value
     * @return boolean:returns true if event list supports this field
     */
    private boolean isSupportedField(int type, String value) {
        return list.isSupportedField(type);
    }

    /**
     * This method creates a hashtable with keys and values associated with
     * blackberry eventlist attributes.
     *
     * @param String: eventString
     * @return Hashtable
     */
    private Hashtable buildMapFromXML(String eventString) {
        int startMarkerIndex = eventString.indexOf(START_MARKER);
        int endMarkerIndex = eventString.indexOf(END_MARKER);

        if(startMarkerIndex == -1 || endMarkerIndex == -1) {
            Dialog.inform("Improper data from the server.");
            return null;
        }

        Hashtable eventMap   = new Hashtable();
        String startTag        = "";
        String endTag          = "";
        String degeneratedTag  = "";
        String data            = "";

        for (int i = 0; i < tagArray.length; i++) {
            String tag = tagArray[i];

            startTag = "<"+tag+">";

            int startTagIndex = eventString.indexOf(startTag);
            if(startTagIndex == -1) {
                continue;
            }

            degeneratedTag = "<"+tag+"/>";
            int degeneratedTagIndex = eventString.indexOf(degeneratedTag);
            if(degeneratedTagIndex != -1) {
                continue;
            }

            endTag = "</"+tag+">";

            int endTagIndex = eventString.indexOf(endTag);
            if(endTagIndex != -1) {
                data = eventString.substring(startTagIndex+startTag.length(), endTagIndex);
            } else {
                data = "";
            }

            eventMap.put(tag, data);
        }
        return eventMap;
    }

    /**
     * Get time from date in "yyyyMMddTHHmmss" String format
     * @param field date in "yyyyMMddTHHmmss" String format
     * @return time from date in "yyyyMMddTHHmmss" String format
     */
    private long getDate(String field) {

        int  day    = 0 ;
        int  mounth = 0 ;
        int  year   = 0 ;
        int  hour   = 0 ;
        int  minute = 0 ;
        int  second = 0 ;

        long offset = 0 ;

        Calendar         date  = null ;
        StaticDataHelper sdh   = null ;

        sdh = new StaticDataHelper();

        if (sdh.enableAdjustDate()) {
            offset = getTimeZoneOffSet();
        } else {
            offset = 0;
        }

        day    = Integer.parseInt (field.substring( 6  , 8  ));
        mounth = Integer.parseInt (field.substring( 4  , 6  ));
        year   = Integer.parseInt (field.substring( 0  , 4  ));
        hour   = Integer.parseInt (field.substring( 9  , 11 ));
        minute = Integer.parseInt (field.substring( 11 , 13 ));
        second = Integer.parseInt (field.substring( 13 , 15 ));

        date = Calendar.getInstance();

        date.set(Calendar.DAY_OF_MONTH , day        ) ;
        date.set(Calendar.MONTH        , mounth - 1 ) ;
        date.set(Calendar.YEAR         , year       ) ;
        date.set(Calendar.HOUR_OF_DAY  , hour       ) ;
        date.set(Calendar.MINUTE       , minute     ) ;
        date.set(Calendar.SECOND       , second     ) ;

        return date.getTime().getTime() + offset;

    }

    /**
     * return seconds from minutes in String format
     * @param field minutes in String format
     * @return seconds
     */
    private int getAlarmSecondsBefore(String field)
    throws NumberFormatException {
        int alarmMinutesBefore = Integer.parseInt(field) * 60;
        return alarmMinutesBefore;
    }

    /**
     * return timezone offser
     */
    private long getTimeZoneOffSet() {

        Calendar now = null;
        TimeZone zn = null;
        long offset = 0;

        now = Calendar.getInstance();

        now.setTime(new Date());

        zn = TimeZone.getDefault();

        offset = zn.getOffset(1                ,
                              now.YEAR         ,
                              now.MONTH        ,
                              now.DAY_OF_MONTH ,
                              now.DAY_OF_WEEK  ,
                              now.MILLISECOND  );

        return offset;

    }

}